
NAME = u'PC-BASIC'
PACKAGE = 'pcbasic'
VERSION = u'2.0.2'
COPYRIGHT = u'(C) Copyright 2013--2019 Rob Hagemans.'
LICENCE = u'GPLv3'
AUTHOR = u'Rob Hagemans'
EMAIL = u'robhagemans@yahoo.co.uk'
URL = u'http://pc-basic.org'
DESCRIPTION = u'A free, cross-platform emulator for the GW-BASIC family of interpreters.'
LONG_DESCRIPTION = u'''\
PC-BASIC is a free, cross-platform interpreter for GW-BASIC, Advanced BASIC (BASICA), PCjr Cartridge Basic and Tandy 1000 GWBASIC.
It interprets these BASIC dialects with a high degree of accuracy, aiming for bug-for-bug compatibility.
PC-BASIC emulates the most common video and audio hardware on which these BASICs used to run.
PC-BASIC runs plain-text, tokenised and protected .BAS files.
It implements floating-point arithmetic in the Microsoft Binary Format (MBF) and can therefore
read and write binary data files created by GW-BASIC.'''
CLASSIFIERS = [
    u'Development Status :: 4 - Beta',
    u'Intended Audience :: End Users/Desktop',
    u'Intended Audience :: Developers',
    u'Topic :: System :: Emulators',
    u'Topic :: Software Development :: Interpreters',
    u'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    u'Programming Language :: Python :: 2.7',
    u'Programming Language :: Python :: 3.5',
]
KEYWORDS = [u'emulator', u'interpreter', u'basic', u'retro', u'legacy', u'gwbasic', u'basica', u'pcjr', u'tandy']
