"""
PC-BASIC - inputs package
Keyboard, screen copy, pen and joystick handling

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from .keyboard import Keyboard
from .inputmethods import Pen, Stick, ScreenCopyHandler
